from connexion.lifecycle import ConnexionResponse, ConnexionRequest
from unittest.mock import MagicMock, AsyncMock, patch

from asgimiddlewares.custom_exception import CustomExceptionMiddleware


@patch("asgimiddlewares.custom_exception.LOG")
def test_custom_exception_middleware(mock_log: MagicMock) -> None:
    request = ConnexionRequest({"type": "http"})
    exception = Exception("Test exception")
    mock_app = AsyncMock()

    middleware = CustomExceptionMiddleware(mock_app)

    response = middleware.common_error_handler(request, exception)

    assert isinstance(response, ConnexionResponse)
    assert response.status_code == 500
    assert mock_log.error.called



""" Middleware for handling path """

import logging
from typing import Optional

from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.routing import Router, Mount, Match, Route


LOG = logging.getLogger(__name__)


class PathIdMiddleware:  # pylint: disable=too-few-public-methods
    """
    Update scope with path_id

    Needs to be positioned after CustomRoutingMiddleware
    to have access to the router key provided by
    the middleware.
    """

    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        router: Optional[Router] = scope.get("router")
        if not router:
            raise KeyError(
                "Scope does not contain the 'router' key. "
                "Make sure this middleware is placed right "
                "after CustomRoutingMiddleware!"
            )
        if scope.get("path") and scope.get("type") == "http" and router:
            for mount in router.routes:
                if scope.get("state", {}).get("path_id"):
                    break
                if not isinstance(mount, Mount):
                    continue
                if not isinstance(mount.app, Router):
                    continue

                match, child_scope = mount.matches(scope)
                if match != Match.FULL:
                    continue
                new_scope = {**scope, **child_scope}
                for route in mount.app.routes:
                    if not isinstance(route, Route):
                        continue
                    match, _ = route.matches(new_scope)

                    if match == Match.FULL:
                        scope["state"]["path_id"] = mount.path + route.path_format
                        break
            if not scope.get("state", {}).get("path_id"):
                # Not a part of application API, possibly swagger UI
                scope["state"]["path_id"] = ''
        await self.app(scope, receive, send)



""" Custom middleware to handle exception with extra logging """

import logging


from connexion.exceptions import InternalServerError
from connexion.lifecycle import ConnexionResponse
from connexion.middleware.exceptions import ExceptionMiddleware

from starlette.requests import Request

from .utils import logging_ctx_var

LOG = logging.getLogger(__name__)


class CustomExceptionMiddleware(
    ExceptionMiddleware  # type: ignore
):  # pylint: disable=too-few-public-methods
    """Custom Exception Middleware to handle unhandled exceptions"""

    @staticmethod
    def common_error_handler(_request: Request, exc: Exception) -> ConnexionResponse:
        """Default handler for any unhandled Exception"""
        LOG.error(
            {
                "message": "Unhandled exception occurred",
                "unhandled_exception": True,
            },
            exc_info=exc,
        )
        # this is needed as the unhandled exception are not
        # caught by the error handlers
        internal_server_error = InternalServerError()
        internal_server_error.ext = {"trace_id": logging_ctx_var.get().get("trace_id")}

        return internal_server_error.to_problem()



"""Connexion-based ASGI middleware collection with utilities."""

import enum

from asgimiddlewares.custom_exception import CustomExceptionMiddleware
from asgimiddlewares.custom_header import CustomHeaderMiddleware
from asgimiddlewares.extended_logging import ExtendedLoggingMiddleware
from asgimiddlewares.path_id import PathIdMiddleware
from asgimiddlewares.prometheus import PrometheusMiddleware
from asgimiddlewares.request_time import RequestTimeMiddleware
from asgimiddlewares.routing import CustomRoutingMiddleware
from asgimiddlewares.utils import (
    replace_middleware,
    server_request_hook,
    request_time_ctx_var,
    logging_ctx_var,
)

__all__ = [
    "CustomExceptionMiddleware",
    "CustomHeaderMiddleware",
    "ExtendedLoggingMiddleware",
    "PathIdMiddleware",
    "PrometheusMiddleware",
    "RequestTimeMiddleware",
    "CustomRoutingMiddleware",
    "replace_middleware",
    "server_request_hook",
    "CustomMiddlewarePosition",
    "request_time_ctx_var",
    "logging_ctx_var",
]


class CustomMiddlewarePosition(enum.Enum):
    """
    Custom enum for positioning middlewares. Imitates Connexion's MiddlewarePosition.

    Replace BEFORE_EXCEPTION
    """

    BEFORE_CUSTOM_EXCEPTION = CustomExceptionMiddleware
    BEFORE_ROUTING = CustomRoutingMiddleware



"""Contains improved routing middleware for Connexion/Starlette."""

from typing import Optional, Dict, Any

import starlette
from connexion.middleware.routing import RoutingAPI, RoutingMiddleware
from connexion.spec import Specification
from starlette.routing import Route
from starlette.types import Receive, Scope, Send


class CustomRoutingMiddleware(  # pragma: no cover
    RoutingMiddleware  # type: ignore
):  # pylint: disable=too-few-public-methods
    """
    Adjusted Connexion Routing middleware that use CustomRoutingAPI
    instead of connexion RoutingAPI.
    """

    def add_api(
        self,
        specification: Specification,
        base_path: Optional[str] = None,
        arguments: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:  # pragma: no cover
        """Add an API to the router based on a OpenAPI spec.

        Does exactly what parent func does, but the routes are sorted, because
        starlette iterates over routes and performs regex matching until a match is
        found. This means that more general routes might be matched before the specific
        variants, which is not what we want.

        Example:
        Incoming request on path:
        /registry/registry.com/repository/ubi8/python39/tag/tag1

        /registry/{registry}/repository/{repository}
            ^ if not correctly sorted this will get matched first,
              because repository is a path
        /registry/{registry}/repository/{repository}/tag/{tag}

        :param specification: OpenAPI spec.
        :param base_path: Base path where to add this API.
        :param arguments: Jinja arguments to replace in the spec.
        """
        api = RoutingAPI(
            specification,
            base_path=base_path,
            arguments=arguments,
            next_app=self.app,
            **kwargs,
        )

        routes_len = len(api.router.routes)

        def sorting_func(route: Route) -> int:
            """
            Sorts routing rules to avoid more general rules to be matched first.
            """

            # if path does not have any parameters, put it on top
            if not route.param_convertors:
                return routes_len + 1
            # otherwise use the length of the regex pattern
            return len(route.path_regex.pattern)

        api.router.routes.sort(key=sorting_func, reverse=True)

        # If an API with the same base_path was already registered, chain the new API
        # as its default. This way, if no matching route is found on the first API,
        # the request is forwarded to the new API.
        for route in self.router.routes:
            if (
                isinstance(route, starlette.routing.Mount)
                and route.path == api.base_path
            ):
                route.app.default = api.router  # type: ignore

        self.router.mount(api.base_path, app=api.router)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        scope["router"] = self.router
        return await super().__call__(scope, receive, send)



""" Middleware for handling request time calculation """

import time

from starlette.types import ASGIApp, Receive, Scope, Send, Message

from .utils import request_time_ctx_var


class RequestTimeMiddleware:  # pylint: disable=too-few-public-methods
    """Measure request time and store it in context variable"""

    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        time_ref: float = time.perf_counter()

        async def wrapped_send(message: Message) -> None:
            if message["type"] == "http.response.start":
                request_time = time.perf_counter() - time_ref
                request_time_ctx_var.set(request_time)

            await send(message)

        await self.app(scope, receive, wrapped_send)



from starlette.routing import Route, Mount
from unittest.mock import MagicMock, patch, AsyncMock

import pytest


from asgimiddlewares import CustomRoutingMiddleware


@pytest.mark.parametrize(
    ["routes", "expected_order"],
    [
        pytest.param(
            [Route("/ping", lambda: "pong"), Route("/foo/{foo}", lambda foo: foo)],
            (1, 0),
        ),
        pytest.param(
            [
                Route("/foo", lambda: "foo"),
                Route("/foo/{bar}/{spam}", lambda bar, spam: (bar, spam)),
                Route("/foo/{bar}", lambda bar: bar),
            ],
            (2, 0, 1),
        ),
    ],
)
@patch("asgimiddlewares.routing.RoutingAPI")
def test_routing_middleware(
    mock_routing_api_constructor: MagicMock,
    routes: list[Route],
    expected_order: tuple[int],
):
    mock_app = AsyncMock()
    middleware = CustomRoutingMiddleware(mock_app)
    middleware.router = MagicMock()
    middleware.router.routes = [Mount("/v1", mock_app)]
    mock_api = mock_routing_api_constructor.return_value
    mock_api.router.routes = routes.copy()
    mock_api.base_path = "/v1"

    mock_specification = MagicMock()
    middleware.add_api(mock_specification, "/v1")

    mock_routing_api_constructor.assert_called_once_with(
        mock_specification, base_path="/v1", arguments=None, next_app=mock_app
    )

    # Check if routes are sorted  properly
    for route, index in zip(routes, expected_order):
        assert mock_api.router.routes[index] == route

    assert middleware.router.routes[0].app.default == mock_api.router

    middleware.router.mount.assert_called_once_with("/v1", app=mock_api.router)


@pytest.mark.asyncio
async def test_scope_change():
    middleware = CustomRoutingMiddleware(AsyncMock())
    mock_router = AsyncMock()
    middleware.router = mock_router
    scope = {"type": "http"}
    await middleware(scope, AsyncMock(), AsyncMock())
    assert scope["router"] == mock_router



from unittest.mock import MagicMock

import pytest

from typing import Any

from opentelemetry.trace import INVALID_SPAN, Span

from asgimiddlewares.utils import replace_middleware, server_request_hook


@pytest.mark.parametrize(
    ["middleware_list", "original", "replacement", "expected"],
    [
        pytest.param(
            [1, 2, 3],
            2,
            4,
            [1, 4, 3],
            id="replace middle",
        )
    ],
)
def test_replace_middleware(
    middleware_list: list[Any], original: Any, replacement: Any, expected: Any
) -> None:
    """
    Find a specific class amongst a list of classes, and replace it.
    """
    replace_middleware(middleware_list, original, replacement)
    assert middleware_list == expected


def test_replace_middleware_exc() -> None:
    """
    Find a specific class amongst a list of classes, and replace it.
    """
    with pytest.raises(ValueError):
        replace_middleware([1, 2, 3], 5, 4)


@pytest.mark.parametrize(["span"], [(None,), (INVALID_SPAN,)])
def test_server_request_hook_with_no_span(span: Span) -> None:
    scope: dict[str, Any] = {"state": {}}
    server_request_hook(span, scope)
    assert "trace_id" not in scope["state"]


def test_server_request_hook() -> None:
    span = MagicMock()
    span.is_recording.return_value = True
    span.get_span_context.return_value.trace_id = 1234567890
    scope: dict[str, Any] = {"state": {}}
    server_request_hook(span, scope)
    assert scope["state"]["trace_id"] == "0x499602d2"



import pytest

from unittest.mock import AsyncMock, MagicMock

from starlette.routing import Route, Mount, WebSocket, Router

from asgimiddlewares import PathIdMiddleware


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["path", "expected"],
    [
        ("/v1/ping", "/v1/ping"),
        ("/v1/foo/bar", "/v1/foo/{foo}"),
        ("/v2/ping", "/v2/ping"),
        ("/v1/ui/", ""),
    ],
)
async def test_path_id_middleware(path: str, expected: str):
    middleware = PathIdMiddleware(AsyncMock())

    mock_router = AsyncMock()

    mock_socket = WebSocket({"type": "websocket"}, MagicMock(), MagicMock())

    mock_mount0 = Mount("/v0", MagicMock())
    mock_mount1 = Mount("/v1", MagicMock(spec=Router))
    mock_mount2 = Mount("/v2", MagicMock(spec=Router))
    mock_router.routes = [mock_socket, mock_mount0, mock_mount1, mock_mount2]
    route1 = Route("/ping", lambda: "pong")
    route2 = Route("/foo/{foo}", lambda foo: f"bar: {foo}")

    mock_mount1.app.routes = [route1, route2]
    mock_mount2.app.routes = [mock_socket, route1]

    mock_scope = {
        "router": mock_router,
        "path": path,
        "type": "http",
        "state": {},
        "method": "GET",
    }
    await middleware(mock_scope, AsyncMock(), AsyncMock())
    assert mock_scope["state"]["path_id"] == expected


@pytest.mark.asyncio
async def test_middleware_exception():
    mock_scope = {
        "path": "/path",
        "type": "http",
        "state": {},
        "method": "GET",
    }
    middleware = PathIdMiddleware(AsyncMock())
    with pytest.raises(KeyError):
        await middleware(mock_scope, AsyncMock(), AsyncMock())



""" Middleware for handling extended logging """

from typing import Dict, List, Tuple, Iterable

from starlette.datastructures import Headers
from starlette.types import ASGIApp, Receive, Scope, Send, Message

from .utils import logging_ctx_var


_FIELD_MAPPING = {
    "method": lambda scope, headers: scope.get("method"),
    "path": lambda scope, headers: scope.get("path"),
    "path_id": lambda scope, headers: None,  # handled in response
    "protocol": lambda scope, headers: f"{scope.get('type')}"
    f"/{scope.get('http_version')}",
    "query": lambda scope, headers: scope.get("query_string", b"").decode("utf-8"),
    "referer": lambda scope, headers: headers.get("referer"),
    "remote_address": lambda scope, headers: scope.get("client", ("-", "-"))[0],
    "response_length": lambda scope, headers: None,  # handled in response
    "status": lambda scope, headers: None,  # handled in response
    "username": lambda scope, headers: "-",  # TODO
    "user_agent": lambda scope, headers: headers.get("user-agent"),
    "trace_id": lambda scope, headers: scope["state"].get("trace_id", "-"),
    "True-Client-IP": lambda scope, headers: headers.get("true-client-ip"),
    "X-Akamai-RH-Edge-Id": lambda scope, headers: headers.get("x-rh-edge-request-id"),
    "X-Forwarded-For": lambda scope, headers: headers.get("x-forwarded-for"),
    "X-Forwarded-Proto": lambda scope, headers: headers.get("x-forwarded-proto"),
    "X-Forwarded-Port": lambda scope, headers: headers.get("x-forwarded-port"),
    "X-Forwarded-Host": lambda scope, headers: headers.get("x-forwarded-host"),
}
POSSIBLE_FIELDS = list(_FIELD_MAPPING.keys())


DEFAULT_SETTINGS = tuple(POSSIBLE_FIELDS)


def parse_headers(headers: List[Tuple[bytes, bytes]]) -> Dict[str, str]:
    """
    Parse scope headers into a dictionary

    Args:
        headers (List[bytes, bytes]]): List of headers

    Returns:
        Dict[str, str]: Dictionary of headers
    """
    return {pair[0].decode("utf-8"): pair[1].decode("utf-8") for pair in headers}


class ExtendedLoggingMiddleware:  # pylint: disable=too-few-public-methods
    """
    Custom middleware to make selected request related
    variables available outside of the request loop.
    """

    def __init__(self, app: ASGIApp, fields: Iterable[str] = DEFAULT_SETTINGS) -> None:
        """
        To override default fields, use functools.partial() with the required
        keyword argument.

        :param app: ASGI application or next middleware.
                    Is called in the middleware stack.
        :param fields: Fields to be included in the logging.
                       See POSSIBLE_FIELDS for possible fields.
        """
        for field in fields:
            if field not in _FIELD_MAPPING:
                raise ValueError(f"Unknown field to log: '{field}'!")
        self.app = app
        self.fields = fields

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope_headers = parse_headers(scope.get("headers", []))

        # base uvicorn.access log structure
        # values from scope, send.message or headers
        data = {key: _FIELD_MAPPING[key](scope, scope_headers) for key in self.fields}

        async def send_wrapper(message: Message) -> None:
            if message["type"] == "http.response.start":
                # Fill in fields that require to be filled
                # after the request is processed
                if "status" in data:
                    data["status"] = message["status"]
                if "path_id" in data:
                    data["path_id"] = scope.get("state", {}).get("path_id")
                headers = Headers(scope=message)
                if "response_length" in data:
                    data["response_length"] = int(headers.get("content-length", 0))

            await send(message)

        logging_ctx_var.set(data)

        await self.app(scope, receive, send_wrapper)



import pytest

from unittest.mock import AsyncMock

from asgimiddlewares import CustomHeaderMiddleware


@pytest.mark.asyncio
async def test_custom_header_middleware_with_csp() -> None:
    mock_app = AsyncMock()
    middleware = CustomHeaderMiddleware(mock_app)
    scope = {"type": "http", "path": "/some-path", "state": {"trace_id": "123456"}}
    mock_receive = AsyncMock()
    mock_send = AsyncMock()

    await middleware(scope, mock_receive, mock_send)

    _, __, send_with_extra_headers_call = mock_app.call_args.args
    assert send_with_extra_headers_call is not None

    await send_with_extra_headers_call(
        {"type": "http.response.start", "status": 200, "headers": []}
    )

    mock_send.assert_awaited_once_with(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"trace_id", b"123456"),
                (
                    b"content-security-policy",
                    b"default-src 'none'; frame-ancestors 'none'",
                ),
            ],
        }
    )


@pytest.mark.parametrize(
    "request_path",
    [
        "/ui",
        "/v1/ui/#/Certification projects/graphql.cert_projects.replace_certification_project",
    ],
)
@pytest.mark.asyncio
async def test_custom_header_middleware_without_csp(request_path: str) -> None:
    mock_app = AsyncMock()
    middleware = CustomHeaderMiddleware(mock_app)
    scope = {"type": "http", "path": request_path, "state": {"trace_id": "123456"}}
    mock_receive = AsyncMock()
    mock_send = AsyncMock()

    await middleware(scope, mock_receive, mock_send)

    _, __, send_with_extra_headers_call = mock_app.call_args.args
    assert send_with_extra_headers_call is not None

    await send_with_extra_headers_call(
        {"type": "http.response.start", "status": 200, "headers": []}
    )
    mock_send.assert_awaited_once_with(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"trace_id", b"123456"),
            ],
        }
    )


@pytest.mark.asyncio
async def test_custom_header_middleware_websocket() -> None:
    mock_app = AsyncMock()
    middleware = CustomHeaderMiddleware(mock_app)
    scope = {"type": "websocket", "path": "/some-path", "state": {"trace_id": "123456"}}

    await middleware(scope, AsyncMock(), AsyncMock())
    mock_app.assert_called_once()



""" Middleware for working with response headers """

from typing import Iterable

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Receive, Scope, Send, Message


class CustomHeaderMiddleware:  # pylint: disable=too-few-public-methods
    """
    Custom ASGI Header Middleware

    Extends header with trace_id and CSP
    """

    def __init__(
        self, app: ASGIApp, csp_disable: Iterable[str] = ("/ui", "/v1/ui")
    ) -> None:
        """
        To override default base_paths, use functools.partial() with the required
        kwargs.

        :param Iterable[str] csp_disable: Specify all paths that should not receive the
        CSP header. ALL paths that START with any of the strings from this iterable are
        disabled. That means string '/v1/ui' disables also path '/v1/ui/#'.
        """
        self.app = app
        self.csp_disable = set(csp_disable)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        async def send_with_extra_headers(message: Message) -> None:
            if message.get("type") == "http.response.start":
                headers = MutableHeaders(scope=message)
                headers.append("trace_id", scope["state"].get("trace_id", "-"))

                path = scope.get("path", "")
                # handle CSP
                if all(not path.startswith(disable) for disable in self.csp_disable):
                    headers.append(
                        "Content-Security-Policy",
                        "default-src 'none'; frame-ancestors 'none'",
                    )

            await send(message)

        await self.app(scope, receive, send_with_extra_headers)



from typing import Iterable, Dict, Any, Optional
from unittest.mock import patch, MagicMock, AsyncMock

import pytest

from asgimiddlewares.prometheus import (
    _setup_prometheus,
    PrometheusMiddleware,
)


@patch("asgimiddlewares.prometheus.start_http_server")
@patch("asgimiddlewares.prometheus.multiprocess.MultiProcessCollector")
@patch("asgimiddlewares.prometheus.CollectorRegistry")
@pytest.mark.parametrize(["port"], [(1,), ("1",)])
def test_setup_prometheus(
    mock_registry: MagicMock,
    mock_multiprocess_collector: MagicMock,
    mock_start_server: MagicMock,
    port: Any,
):
    _setup_prometheus(port)
    mock_start_server.assert_called_once_with(port)
    mock_registry.assert_called_once()
    mock_multiprocess_collector.assert_called_once_with(mock_registry())


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["excluded_paths", "scope", "endpoint_for_metrics"],
    [
        (
            ["/v1/ping"],
            {"path": "/v1/sample", "method": "GET", "state": {"path_id": "/v1/sample"}},
            "/v1/sample",
        ),
        (["/v1/ping"], {"path": "/v1/ping", "method": "GET", "path_params": {}}, None),
        (
            [],
            {
                "path": "/v1/sample/id/123",
                "method": "GET",
                "state": {"path_id": "/v1/sample/id/<identifier>"},
            },
            "/v1/sample/id/<identifier>",
        ),
        (
            [],
            {
                "path": "/v1/sample/id/123/other/456",
                "method": "GET",
                "state": {"path_id": "/v1/sample/id/<identifierA>/other/<identifierB>"},
            },
            "/v1/sample/id/<identifierA>/other/<identifierB>",
        ),
    ],
)
@patch(
    "asgimiddlewares.prometheus.time.perf_counter",
    MagicMock(return_value=0),
)
@patch("asgimiddlewares.prometheus.Counter")
@patch("asgimiddlewares.prometheus.Histogram")
@patch("asgimiddlewares.prometheus.disable_created_metrics")
@patch("asgimiddlewares.prometheus.prometheus_client.REGISTRY.unregister")
@patch("asgimiddlewares.prometheus._setup_prometheus")
async def test_prometheus_middleware(
    mock_setup_prometheus: MagicMock,
    mock_unregister: MagicMock,
    mock_disable_create: MagicMock,
    mock_histogram_constructor: MagicMock,
    mock_counter_constructor: MagicMock,
    excluded_paths: Iterable[str],
    scope: Dict[str, Any],
    endpoint_for_metrics: Optional[str],
) -> None:
    # ASGI app is a callable
    mock_app = AsyncMock()

    # This step would normally be done inside Connexion
    middleware = PrometheusMiddleware(
        mock_app, service_name="foo", port=5002, excluded_paths=excluded_paths
    )
    middleware.hostname = "Marvin"
    mock_setup_prometheus.assert_called_once_with(5002)
    mock_counter_constructor.assert_called_once()
    mock_histogram_constructor.assert_called_once()
    # The middleware uses a set, duplicates will be deleted
    for excluded_path in excluded_paths:
        assert excluded_path in middleware.excluded_paths
    # Check for function called in the constructor to reduce metrics
    mock_disable_create.assert_called_once()
    mock_unregister.assert_called()

    # Call the middleware, simulates Connexion's stack.
    # This is called when a request is made.
    await middleware(scope, AsyncMock(), AsyncMock())

    mock_app.assert_called_once()
    # This is a simple hack to call the "send" function
    # which would otherwise be called in the middleware stack.
    _, __, send = mock_app.call_args.args
    await send({"type": "http.response.start", "status": 200})

    if endpoint_for_metrics is not None:
        method = scope["method"]
        # Check if correct labels were applied
        middleware.counter.labels.assert_called_once_with("Marvin", "200", method)
        middleware.histogram.labels.assert_called_once_with(
            "Marvin", "200", method, endpoint_for_metrics
        )

        # Check if correct call was made on the labels
        middleware.counter.labels.return_value.inc.assert_called_once_with(1)
        middleware.histogram.labels.return_value.observe.assert_called_once_with(0)
    else:
        # Nothing was incremented if the endpoint is supposed to be ignored
        middleware.counter.labels.assert_not_called()
        middleware.histogram.labels.assert_not_called()

        middleware.counter.labels.return_value.inc.assert_not_called()
        middleware.histogram.labels.return_value.observe.assert_not_called()



import pytest

from unittest.mock import AsyncMock

from asgimiddlewares.utils import request_time_ctx_var
from asgimiddlewares import RequestTimeMiddleware


@pytest.mark.asyncio
async def test_request_time_middleware() -> None:
    mock_app = AsyncMock()
    middleware = RequestTimeMiddleware(mock_app)

    await middleware({}, AsyncMock(), AsyncMock())

    mock_app.assert_called_once()

    _, __, send = mock_app.call_args.args
    await send({"type": "http.response.start", "status": 200})

    assert request_time_ctx_var.get() is not None
    assert isinstance(request_time_ctx_var.get(), float)



"""Utilities for middleware placement and context handling."""

from typing import Any, Optional
from contextvars import ContextVar

from opentelemetry.trace import Span, INVALID_SPAN
from starlette.types import Scope

logging_ctx_var: ContextVar[dict[str, Any]] = ContextVar("extended_logs", default={})

request_time_ctx_var: ContextVar[Optional[float]] = ContextVar(
    "request_time", default=None
)


def replace_middleware(
    middleware_list: list[Any], original: Any, replacement: Any
) -> None:
    """
    Find a specific class amongst a list of classes, and replace it.
    """
    for i, middleware in enumerate(middleware_list):
        if middleware == original:
            middleware_list.pop(i)
            middleware_list.insert(i, replacement)
            return
    raise ValueError(f"Middleware {original} not found in the list of middlewares")


def server_request_hook(span: Span, scope: Scope) -> None:
    """Request hook to store trace_id to the scope"""
    # We need to check even unrecorded spans.
    # Unrecorded spans can be set by traceparent header
    # with unsampled flag, see https://www.w3.org/TR/trace-context/#traceparent-header
    # but trace_id is also used for debugging purposes
    if span and span != INVALID_SPAN:
        trace_id = None
        span_context = span.get_span_context()
        if span_context.trace_id != 0:
            # trace_id 0 means invalid span
            trace_id = str(hex(span_context.trace_id))
        scope["state"]["trace_id"] = trace_id



import pytest

from typing import Any, Dict
from unittest.mock import MagicMock, AsyncMock, patch

from asgimiddlewares.utils import logging_ctx_var
from asgimiddlewares.extended_logging import (
    parse_headers,
    ExtendedLoggingMiddleware,
)


def test_parse_headers_empty() -> None:
    headers = []
    result = parse_headers(headers)
    assert result == {}


def test_parse_headers_multiple_headers() -> None:
    headers = [
        (b"Content-Type", b"application/json"),
        (b"Authorization", b"Bearer token"),
        (b"X-Request-ID", b"1234567890"),
    ]
    result = parse_headers(headers)
    assert result == {
        "Content-Type": "application/json",
        "Authorization": "Bearer token",
        "X-Request-ID": "1234567890",
    }


@pytest.mark.parametrize(
    ("scope", "message"),
    [
        pytest.param(
            {
                "method": "GET",
                "path": "/v1/path/a/to/b/somewhere/c",
                "path_params": {"a": "a", "b": "b", "c": "c"},
                "type": "http",
                "http_version": "1.1",
                "query_string": b"test",
                "state": {
                    "trace_id": "0x5d66f33f56b69ead5d79f7a928fde097",
                    "path_id": "/v1/path/<a>/to/<b>/somewhere/<c>",
                },
                "client": ("a.b.c.d", 12345),
                "headers": [
                    (b"referer", b"john_doe"),
                    (b"user-agent", b"curl/7.76.1"),
                    (b"true-client-ip", b"a.b.c.d"),
                    (b"x-rh-edge-request-id", b"id"),
                    (b"x-forwarded-for", b"for"),
                    (b"x-forwarded-proto", b"proto"),
                    (b"x-forwarded-port", b"port"),
                    (b"x-forwarded-host", b"host"),
                ],
            },
            {"type": "http.response.start", "status": 200},
        )
    ],
)
@pytest.mark.parametrize(
    ("headers", "expected", "tracked_fields"),
    [
        pytest.param(
            {"content-length": b"123"},
            {
                "method": "GET",
                "path": "/v1/path/a/to/b/somewhere/c",
                "path_id": "/v1/path/<a>/to/<b>/somewhere/<c>",
                "protocol": "http/1.1",
                "query": "test",
                "referer": "john_doe",
                "remote_address": "a.b.c.d",
                "response_length": 123,
                "status": 200,
                "username": "-",
                "user_agent": "curl/7.76.1",
                "trace_id": "0x5d66f33f56b69ead5d79f7a928fde097",
                "True-Client-IP": "a.b.c.d",
                "X-Akamai-RH-Edge-Id": "id",
                "X-Forwarded-For": "for",
                "X-Forwarded-Proto": "proto",
                "X-Forwarded-Port": "port",
                "X-Forwarded-Host": "host",
            },
            None,
            id="content-length present",
        ),
        pytest.param(
            {},
            {
                "method": "GET",
                "path": "/v1/path/a/to/b/somewhere/c",
                "path_id": "/v1/path/<a>/to/<b>/somewhere/<c>",
                "protocol": "http/1.1",
                "query": "test",
                "referer": "john_doe",
                "remote_address": "a.b.c.d",
                "response_length": 0,
                "status": 200,
                "username": "-",
                "user_agent": "curl/7.76.1",
                "trace_id": "0x5d66f33f56b69ead5d79f7a928fde097",
                "True-Client-IP": "a.b.c.d",
                "X-Akamai-RH-Edge-Id": "id",
                "X-Forwarded-For": "for",
                "X-Forwarded-Proto": "proto",
                "X-Forwarded-Port": "port",
                "X-Forwarded-Host": "host",
            },
            None,
            id="content-length missing",
        ),
        pytest.param(
            {},
            {
                "path": "/v1/path/a/to/b/somewhere/c",
                "path_id": "/v1/path/<a>/to/<b>/somewhere/<c>",
                "query": "test",
            },
            ["path", "path_id", "query"],
        ),
    ],
)
@patch("asgimiddlewares.extended_logging.Headers")
@pytest.mark.asyncio
async def test_extended_logging_middleware(
    mock_headers: MagicMock,
    scope: Dict[str, Any],
    message: Dict[str, Any],
    expected: Dict[str, Any],
    tracked_fields: list[str],
    headers: Dict[str, Any],
) -> None:
    mock_app = AsyncMock()
    mock_headers.return_value = headers
    if tracked_fields is not None:
        middleware = ExtendedLoggingMiddleware(mock_app, tracked_fields)
    else:
        middleware = ExtendedLoggingMiddleware(mock_app)

    await middleware(scope, AsyncMock(), AsyncMock())

    mock_app.assert_called_once()

    _, __, send = mock_app.call_args.args
    await send(message)

    assert logging_ctx_var.get() == expected


def test_extended_logging_middleware_invalid_field():
    with pytest.raises(ValueError, match="Unknown field to log: 'foo'!"):
        ExtendedLoggingMiddleware(MagicMock(), ("foo", "boar"))



"""Custom Prometheus middleware designed specifically for
Connexion AsyncApp. It has to be added BEFORE EXCEPTION,
which is by the end of the Middleware stack.

It requires the environment variable PROMETHEUS_MULTIPROC_DIR."""

import os
import time
from typing import Iterable, Any

import prometheus_client

from prometheus_client import (
    Counter,
    Histogram,
    disable_created_metrics,
    PROCESS_COLLECTOR,
    PLATFORM_COLLECTOR,
    GC_COLLECTOR,
    CollectorRegistry,
    multiprocess,
    start_http_server,
)
from starlette.types import ASGIApp, Scope, Receive, Send


def _setup_prometheus(port: int) -> None:
    """Starts an HTTP prometheus multiprocessing server at the specified port."""
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    start_http_server(port)


__all__ = ["PrometheusMiddleware"]


# pylint: disable=too-few-public-methods
class PrometheusMiddleware:
    """Connexion Middleware class for Prometheus metrics exposure."""

    def __init__(
        self,
        app: ASGIApp,
        service_name: str = "",
        port: int = 5001,
        excluded_paths: Iterable[str] = ("",),
    ):
        """
        This constructor is intended to be called inside a Connexion
        middleware stack and not by the user. To add it to the stack,
        use app.add_middleware(PrometheusMiddleware). If you wish to
        specify non_default arguments, use functools.partial() to pass
        them as kwargs.

        :param ASGIApp app: ASGI app or a middleware layer.
        :param str service_name: Name of the service to appear in Prometheus metrics,
        defaults to empty string.
        :param int port: Port to expose the metrics, defaults to 5001.
        :param Iterable[str] excluded_paths: Endpoints to exclude from metrics.
        If you wish to exclude an endpoint containing a variable,
        enclose all variables in {curly braces} like so:
        `excluded_paths=("/v1/sample/id/{identifier}",)`
        """
        _setup_prometheus(port)
        if service_name:
            service_name = f"{service_name}_"
        self.counter = Counter(
            f"{service_name}http_request_total",
            "Accesses of HTTP endpoints",
            labelnames=("hostname", "status", "method"),
        )
        self.histogram = Histogram(
            f"{service_name}http_request_duration_seconds",
            "HTTP request duration in seconds",
            labelnames=("hostname", "status", "method", "url_rule"),
        )
        self.hostname = os.environ.get("HOSTNAME", "localhost")

        self.app: ASGIApp = app
        # Disable measuring the UNIX time when a metric was created
        disable_created_metrics()
        # Disable metrics related to the host
        for collector in (PROCESS_COLLECTOR, PLATFORM_COLLECTOR, GC_COLLECTOR):
            prometheus_client.REGISTRY.unregister(collector)
        self.excluded_paths = set(excluded_paths)

    async def _timed_call(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Wrap the request execution to measure the time of execution.
        :param Scope scope: Mapping passed along with the ASGI request.
        :param Receive receive: Callable object for request manipulation.
        :param Send send: Callable object for response manipulation.
        :return: None
        """
        method = scope["method"]
        time_ref = time.perf_counter()

        async def wrapped_send(response: Any) -> None:
            """
            Wrapped send function to measure the metrics.
            :param response: Mapping representing an HTTP response.
            :return: Calls an ASGI `send` function and returns its value.
            """
            # Measure time upon receiving the response head
            if response["type"] == "http.response.start":
                status_code = str(response["status"])
                # get path_id prepared by the PathIdMiddleware
                path_id = scope.get("state", {}).get("path_id")
                self.histogram.labels(
                    self.hostname, status_code, method, path_id
                ).observe(time.perf_counter() - time_ref)
                self.counter.labels(self.hostname, status_code, method).inc(1)
            await send(response)

        await self.app(scope, receive, wrapped_send)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Use the middleware. Called inside the connexion middleware stack.
        :param Scope scope: Mapping passed along with the ASGI request.
        :param Receive receive: Callable object for request manipulation.
        :param Send send: Callable object for response manipulation.
        """
        if "path" in scope:
            if scope["path"] not in self.excluded_paths:
                await self._timed_call(scope, receive, send)
                return
        await self.app(scope, receive, send)
